'------------------------------------------------------------------------------------
' PBJPEG.BAS - PowerBasic JPEG Encoder
'------------------------------------------------------------------------------------
'
#IF NOT %DEF(%WINAPI)
#INCLUDE "win32api.inc"                'add win32api if not already added
#ENDIF

%NOMSG = 0  'Change to 0 to display error messages for debuging or 1 to not show messages

'------------------------------------------------------------------------------------
' JPEG Marker Constants
'------------------------------------------------------------------------------------
' Non-Differential Huffman Coding
%SOF0  = &HC0&    ' Baseline DCT
%SOF1  = &HC1&    ' Extended sequential DCT
%SOF2  = &HC2&    ' Progressive DCT
%SOF3  = &HC3&    ' Spatial (sequential) lossless

' Differential Huffman coding
%SOF5  = &HC5&    ' Differential sequential DCT
%SOF6  = &HC6&    ' Differential progressive DCT
%SOF7  = &HC7&    ' Differential spatial

' Non-Differential arithmetic coding
%JPG    = &HC8&    ' Reserved for JPEG extentions
%SOF9  = &HC9&    ' Extended sequential DCT
%SOF10  = &HCA&    ' Progressive DCT
%SOF11  = &HCB&    ' Spatial (sequential) lossless

' Differential arithmetic coding
%SOF13  = &HCD&    ' Differential sequential DCT
%SOF14  = &HCE&    ' Differential progressive DCT
%SOF15  = &HCF&    ' Differential Spatial

' Other Markers
%DHT    = &HC4&    ' Define Huffman tables
%DAC    = &HCC&    ' Define arithmetic coding conditioning(s)
%RSTM  = &HD0&    ' Restart with modulo 8 count "m"
%RSTM2  = &HD7&    '  to Restart with modulo 8 count "m"
%SOI    = &HD8&    ' Start of image
%EOI    = &HD9&    ' End of image
%SOS    = &HDA&    ' Start of scan
%DQT    = &HDB&    ' Define quantization table(s)
%DNL    = &HDC&    ' Define number of lines
%DRI    = &HDD&    ' Define restart interval
%DHP    = &HDE&    ' Define hierarchical progression
%VEXP  = &HDF&    ' Expand reference components
%APP0  = &HE0&    ' Reserved for application segments
%APPF  = &HEF&    '  to Reserved for application segments
%JPGN  = &HF0&    ' Reserved for JPEG Extentions
%JPGN2  = &HFD&    '  to Reserved for JPEG Extentions
%COM    = &HFE&    ' Comment
%RESM  = &H2&      ' Reserved
%RESM2  = &HBF&    '  to Reserved
%TEM    = &H1&      ' For temporary use in arithmetic coding


%TWOE00 = &H1&
%TWOE01 = &H2&
%TWOE02 = &H4&
%TWOE03 = &H8&
%TWOE04 = &H10&
%TWOE05 = &H20&
%TWOE06 = &H40&
%TWOE07 = &H80&
%TWOE08 = &H100&
%TWOE09 = &H200&
%TWOE10 = &H400&
%TWOE11 = &H800&
%TWOE12 = &H1000&
%TWOE13 = &H2000&
%TWOE14 = &H4000&
%TWOE15 = &H8000&



' Consider these arrays of constants.
' They are initialized with the module and do not change.
GLOBAL g_QLumin()      AS BYTE      ' Standard Luminance Quantum (for 50% quality)
GLOBAL g_QChrom()      AS BYTE      ' Standard Chrominance Quantum (for 50% quality)
GLOBAL g_TwoP()        AS LONG      ' Powers of two 2^P
GLOBAL g_TwoR()        AS LONG      ' Powers of two 2^(R-1)
GLOBAL g_TwoB()        AS BYTE      ' The number of bits in value
GLOBAL g_EIDCTScale()  AS LONG      ' Encoding Constants for scaling IntDCT Coefficients
GLOBAL g_DIDCTScale()  AS LONG      ' Decoding Constants for scaling IntDCT Coefficients
GLOBAL g_ZigZag()      AS LONG      ' Zig Zag order of 8X8 block of samples

' Custom variable types used for this JPEG encoding implementation
TYPE QUANTIZATIONTABLE
  Qk(63)        AS INTEGER ' Quantization Values
  FScale(63)    AS LONG    ' Multiplication values to scale and Quantize FDCT output
  IScale(63)    AS LONG    ' Multiplication values to scale and DeQuantize IDCT input
END TYPE

TYPE HUFFMANTABLE
  vBits(15)    AS BYTE    ' Number of huffman codes of length i+1
  HuffVal(255)  AS BYTE    ' Huffman symbol values
  EHufLN(255)  AS BYTE    ' Huffman codelength in Bits for symbol i
  EHufSI(255)  AS LONG    ' Huffman code size for symbol i
  EHufCO(255)  AS LONG    ' Huffman code for symbol i
  MinCode(15)  AS LONG    '
  MaxCode(15)  AS LONG    ' Largest code value for length i+1
END TYPE

TYPE vComponent
  Ci            AS LONG        ' Component ID                [0-255]
  vHi          AS LONG        ' Horizontal Sampling Factor  [1-4]
  Vi            AS LONG        ' Vertical  Sampling Factor  [1-4]
  Tqi          AS LONG        ' Quantization Table Select  [0-3]
  ptrData      AS INTEGER PTR  ' A dirty workaround to the PB UDT dynamic array issue
END TYPE


'------------------------------------------------------------------------------------
' Global variables
'------------------------------------------------------------------------------------
GLOBAL g_PP      AS LONG    ' Sample Precision [8, 12]
GLOBAL g_YY      AS LONG    ' Number of lines            [Image Height] after clipping
GLOBAL g_XX      AS LONG    ' Number of samples per line  [Image Width]  after clipping
GLOBAL g_Nf      AS LONG    ' Number of components in Frame

GLOBAL g_HMax    AS LONG    ' Maximum horizontal sampling frequency
GLOBAL g_VMax    AS LONG    ' Maximum vertical sampling frequency

GLOBAL g_m_Data() AS BYTE    ' JPEG File Data
GLOBAL g_m_Chr    AS LONG    ' Current Character in g_m_Data
GLOBAL g_m_Ptr    AS LONG    ' Byte index in g_m_Data
GLOBAL g_m_Bit    AS LONG    ' Bit index in g_m_Chr

GLOBAL g_m_Block() AS LONG  ' Buffer for calculating intDCT  (7,7)

GLOBAL g_QTable()    AS QUANTIZATIONTABLE  ' 4 Quantization Tables
GLOBAL g_HuffDC()    AS HUFFMANTABLE      ' 4 DC Huffman Tables
GLOBAL g_HuffAC()    AS HUFFMANTABLE      ' 4 AC Huffman Tables
GLOBAL g_Comp()      AS vComponent        ' Scan Components

GLOBAL g_CompData0() AS INTEGER            ' DCT Coefficients [DC(0),AC(1-63)]*nBlocks
GLOBAL g_CompData1() AS INTEGER
GLOBAL g_CompData2() AS INTEGER

GLOBAL g_m_Quality  AS LONG
GLOBAL g_m_Comment  AS STRING


'------------------------------------------------------------------------------------
' Function prototypes
'------------------------------------------------------------------------------------
DECLARE FUNCTION piaUBOUND(BYREF iArray() AS INTEGER) AS LONG
DECLARE SUB piaREDIM(BYREF iArray() AS INTEGER, BYVAL lValue AS LONG)
DECLARE SUB piaREDIMPRESERVE(BYREF iArray() AS INTEGER, BYVAL lValue AS LONG)
DECLARE SUB piaSET(BYREF iArray() AS INTEGER, BYVAL lNdx AS LONG, BYVAL sValue AS INTEGER)
DECLARE FUNCTION piaGET(BYREF iArray() AS INTEGER, BYVAL lNdx AS LONG) AS INTEGER
DECLARE SUB intDCT()
DECLARE SUB TransformAndQuantize(BYREF vPixel() AS BYTE, BYVAL CompIndex AS LONG, BYREF FScale() AS LONG, _
BYVAL vWidth AS LONG, BYVAL Height AS LONG, BYVAL RowMod AS LONG)
DECLARE FUNCTION PBJPEG_SetQuality(BYVAL vQuality AS LONG) AS LONG
DECLARE SUB ExpandDQT(BYVAL Tqi AS LONG)
DECLARE SUB BuildHuffman(BYREF TheHuff AS HUFFMANTABLE, BYREF Freq() AS LONG)
DECLARE SUB ExpandHuffman(BYREF TheHuff AS HUFFMANTABLE, BYVAL MaxSymbol AS LONG)
DECLARE FUNCTION CodeLength(BYREF TheHuff AS HUFFMANTABLE, BYREF Freq() AS LONG) AS LONG
DECLARE SUB WriteBitsBegin()
DECLARE SUB WriteBitsEnd()
DECLARE SUB WriteBits(BYVAL mask AS LONG, BYVAL CODE AS LONG)
DECLARE SUB EncodeCoefficients(BYREF vData() AS INTEGER, BYREF p AS LONG, BYREF Pred AS LONG, _
BYVAL Td AS LONG, BYVAL Ta AS LONG)
DECLARE SUB CollectStatisticsAC(BYREF vData() AS INTEGER, BYREF FreqAC() AS LONG)
DECLARE SUB CollectStatisticsDCNonInterleaved(BYREF vData() AS INTEGER, BYREF FreqDC() AS LONG)
DECLARE SUB CollectStatisticsDCInterleaved(BYREF vData() AS INTEGER, BYREF FreqDC() AS LONG, _
BYVAL vHi AS LONG, BYVAL Vi AS LONG)
DECLARE SUB PBJPEG_SetSamplingFrequencies(BYVAL H1 AS LONG, BYVAL v1 AS LONG, _
BYVAL H2 AS LONG, BYVAL v2 AS LONG, _
BYVAL H3 AS LONG, BYVAL v3 AS LONG)
DECLARE FUNCTION PBJPEG_SampleHDC(BYVAL lHDC AS LONG, _
BYVAL lWidth AS LONG, _
BYVAL lHeight AS LONG, _
BYVAL lSrcLeft AS LONG, _
BYVAL lSrcTop AS LONG) AS LONG
DECLARE FUNCTION PBJPEG_AddComment(BYVAL vComment AS STRING) AS LONG
DECLARE SUB InsertJFIF()
DECLARE SUB InsertSOF(BYVAL SOFMarker AS LONG)
DECLARE SUB InsertCOM(BYVAL TheComment AS STRING)
DECLARE SUB InsertDQT(BYVAL MarkerPos AS LONG, BYVAL Tqi AS LONG)
DECLARE SUB InsertDHT(BYVAL MarkerPos AS LONG, BYVAL HIndex AS LONG, BYVAL isAC AS LONG)
DECLARE SUB InsertMarker(BYVAL TheMarker AS LONG)
DECLARE SUB InsertSOSNonInterleaved(BYVAL CompIndex AS LONG, BYVAL Td AS LONG, BYVAL Ta AS LONG)
DECLARE SUB InsertSOSInterleaved(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG)
DECLARE SUB InsertSequentialScans(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG)
DECLARE FUNCTION OptimizeHuffmanTables(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG) AS LONG
DECLARE FUNCTION PBJPEG_SaveFile(FileName AS STRING) AS LONG
DECLARE FUNCTION PBJPEG_Initialize() AS LONG
DECLARE SUB Init_BitMasks()
DECLARE SUB GenEIDCT()
DECLARE FUNCTION nBits(BYVAL v AS LONG) AS LONG


'------------------------------------------------------------------------------------
' Pointer-based handling of Integer arrays
'------------------------------------------------------------------------------------
' Based on the work of Stan Durham
FUNCTION piaUBOUND(BYREF iArray() AS INTEGER) AS LONG
  FUNCTION = UBOUND(iArray())
END FUNCTION

'------------------------------------------------------------------------------------

SUB piaREDIM(BYREF iArray() AS INTEGER, BYVAL lValue AS LONG)
  REDIM iArray(lValue)
END SUB

'------------------------------------------------------------------------------------

SUB piaREDIMPRESERVE(BYREF iArray() AS INTEGER, BYVAL lValue AS LONG)
  REDIM PRESERVE iArray(lValue)
END SUB

'------------------------------------------------------------------------------------

SUB piaSET(BYREF iArray() AS INTEGER, BYVAL lNdx AS LONG, BYVAL sValue AS INTEGER)
  iArray(lNdx) = sValue
END SUB

'------------------------------------------------------------------------------------

FUNCTION piaGET(BYREF iArray() AS INTEGER, BYVAL lNdx AS LONG) AS INTEGER
  FUNCTION = iArray(lNdx)
END FUNCTION


'------------------------------------------------------------------------------------
' Integer discrete cosine transformation
'------------------------------------------------------------------------------------
SUB intDCT()
  TRY
      ' Integer Lifted, Loeffler factorisation 22 shifts, 40 adds by Jie Liang, Trac Tran
      ' See IEEE TRANSACTIONS ON SIGNAL PROCESSING VOL49, #12 DEC 2001

      ' This is carefully optimized and its easy to screw its performance up
      ' However its ideally suited to asm, so go for it Vlad!
      LOCAL t0 AS LONG, t1 AS LONG, t2  AS LONG, t3  AS LONG
      LOCAL t4 AS LONG, t5 AS LONG, t6  AS LONG, t7  AS LONG
      LOCAL t8 AS LONG, t9 AS LONG, t10 AS LONG, t11 AS LONG
      LOCAL i  AS LONG

      FOR i = 0 TO 7                                  ' Process 1D intDCT on each row
        t0 = g_m_Block(i, 0)
        t8 = g_m_Block(i, 7)
        t7 = t0 - t8
        t0 = t0 + t8

        t1 = g_m_Block(i, 1)
        t8 = g_m_Block(i, 6)
        t6 = t1 - t8
        t1 = t1 + t8

        t2 = g_m_Block(i, 2)
        t8 = g_m_Block(i, 5)
        t5 = t2 - t8
        t2 = t2 + t8

        t3 = g_m_Block(i, 3)
        t8 = g_m_Block(i, 4)
        t4 = t3 - t8
        t3 = t3 + t8

        ' Top half
        t8 = t0 + t3
        t9 = t1 + t2
        t8 = t8 + t9
        g_m_Block(i, 0) = t8
        g_m_Block(i, 4) = t8 \ 2 - t9                            ' 1/2 *t8

        t10 = t1 - t2
        t11 = t0 - t3
        t10 = (t11 \ 4 + t11 \ 8 + t11 \ 32) - t10              ' 13/32*t11
        g_m_Block(i, 6) = t10
        g_m_Block(i, 2) = t11 - (t10 \ 4 + t10 \ 16 + t10 \ 32)  ' -11/32*t10

        ' Bottom half
        t7 = t7 - (t4 \ 4 + t4 \ 32 + t4 \ 64)                  ' -19/64*t4
        t4 = t4 + (t7 \ 2 + t7 \ 16)                            ' + 9/16*t7
        t7 = t7 - (t4 \ 4 + t4 \ 32 + t4 \ 64)                  ' -19/64*t4

        t6 = t6 - (t5 \ 16 + t5 \ 32)                          ' - 3/32*t5
        t5 = t5 + (t6 \ 8 + t6 \ 16)                            ' + 3/16*t6
        t6 = t6 - (t5 \ 16 + t5 \ 32)                          ' - 3/32*t5

        g_m_Block(i, 3) = t7 - t5
        g_m_Block(i, 5) = t4 - t6

        t8 = t4 + t6
        t11 = t5 + t7 + t8
        g_m_Block(i, 1) = t11
        g_m_Block(i, 7) = t11 \ 2 - t8                            ' 1/2
      NEXT i

      FOR i = 0 TO 7                                  ' Process 1D intDCT on each column
        t0 = g_m_Block(0, i)
        t8 = g_m_Block(7, i)
        t7 = t0 - t8
        t0 = t0 + t8

        t1 = g_m_Block(1, i)
        t8 = g_m_Block(6, i)
        t6 = t1 - t8
        t1 = t1 + t8

        t2 = g_m_Block(2, i)
        t8 = g_m_Block(5, i)
        t5 = t2 - t8
        t2 = t2 + t8

        t3 = g_m_Block(3, i)
        t8 = g_m_Block(4, i)
        t4 = t3 - t8
        t3 = t3 + t8

        ' Top half
        t8 = t0 + t3
        t9 = t1 + t2
        t8 = t8 + t9
        g_m_Block(0, i) = t8
        g_m_Block(4, i) = t8 \ 2 - t9                            ' 1/2 *t8

        t10 = t1 - t2
        t11 = t0 - t3
        t10 = (t11 \ 4 + t11 \ 8 + t11 \ 32) - t10              ' 13/32*t11
        g_m_Block(6, i) = t10
        g_m_Block(2, i) = t11 - (t10 \ 4 + t10 \ 16 + t10 \ 32)  ' -11/32*t10

        ' Bottom half
        t7 = t7 - (t4 \ 4 + t4 \ 32 + t4 \ 64)                  ' -19/64*t4
        t4 = t4 + (t7 \ 2 + t7 \ 16)                            ' + 9/16*t7
        t7 = t7 - (t4 \ 4 + t4 \ 32 + t4 \ 64)                  ' -19/64*t4

        t6 = t6 - (t5 \ 16 + t5 \ 32)                          ' - 3/32*t5
        t5 = t5 + (t6 \ 8 + t6 \ 16)                            ' + 3/16*t6
        t6 = t6 - (t5 \ 16 + t5 \ 32)                          ' - 3/32*t5

        g_m_Block(3, i) = t7 - t5
        g_m_Block(5, i) = t4 - t6

        t8 = t4 + t6
        t11 = t5 + t7 + t8
        g_m_Block(1, i) = t11
        g_m_Block(7, i) = t11 \ 2 - t8                            ' 1/2
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error1",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Quantization
'------------------------------------------------------------------------------------
SUB TransformAndQuantize(BYREF vPixel() AS BYTE, BYVAL CompIndex AS LONG, BYREF FScale() AS LONG, _
  BYVAL vWidth AS LONG, BYVAL Height AS LONG, BYVAL RowMod AS LONG)
  TRY
      LOCAL i        AS LONG  ' Pixel Index (Horizontal)
      LOCAL j        AS LONG  ' Pixel Index (Vertical)
      LOCAL k        AS LONG  ' Index of an 8X8 block of pixels
      LOCAL p        AS LONG  ' DCT Index (horizontal)
      LOCAL q        AS LONG  ' DCT Index (vertical)
      LOCAL t        AS LONG  ' COMPONENT in Zig Zag order at index BlkPtr
      LOCAL n        AS LONG  ' Size of Image
      LOCAL BlkPtr  AS LONG  ' Index for quantized intDCT value (in component data)
      LOCAL BlkMod  AS LONG  ' StepWidth for a Block
      LOCAL RowModS  AS LONG  ' Length of a Sample Row
      LOCAL BlkModS  AS LONG  ' Length of a Sample Row of Blocks
      LOCAL fr      AS LONG
      LOCAL fg      AS LONG
      LOCAL fb      AS LONG
      LOCAL fc      AS LONG

      SELECT CASE CompIndex
        CASE 0  ' Luminance rm = 0.299; gm = 0.587; bm = 0.114; s = 128
            fr = 19595&: fg = 38470: fb = 7471&: fc = 8388608
        CASE 1  ' Chrominance [Blue-Yellow] rm = -0.16874; gm = -0.33126; bm = 0.5; s = 0
            fr = -11059&: fg = -21709&: fb = 32768: fc = 0
        CASE 2  ' Chrominance [Red-Green] rm = 0.5; gm = -0.41869; bm = -0.08131; s=0
            fr = 32768: fg = -27439&: fb = -5329&: fc = 0
      END SELECT

      'REDIM g_Comp(CompIndex).vData(vWidth * Height - 1)
      piaREDIM BYVAL g_Comp(CompIndex).ptrData, vWidth * Height - 1

      ' Reset output Quantized intDCT Coefficient Index
      BlkPtr = 0

      n = Height * RowMod
      RowModS = vWidth * 3
      BlkMod = RowMod - 24
      BlkModS = 8 * RowMod - RowModS
      i = 0

      DO WHILE i < n
        k = RowModS

        DO
            j = i

            ' Get 8X8 block of level shifted YCbCr values
            FOR q = 0 TO 7
              FOR p = 0 TO 7
                  g_m_Block(p, q) = fr * vPixel(i + 2) + fg * vPixel(i + 1) + fb * vPixel(i) - fc
                  i = i + 3
              NEXT p
              i = i + BlkMod
            NEXT q

            ' Take 8X8 block of unscaled DCT coefficients [g_m_Block(0-7, 0-7)],
            ' Scale, Quantize, and store the results in data() array of
            ' COMPONENT in Zig Zag order at index BlkPtr

            ' Calculate the integer DCT (it is scaled by 65536)
            CALL intDCT

            FOR q = 0 TO 7
              FOR p = 0 TO 7
                  t = g_ZigZag(p, q)
                  '                    g_Comp(CompIndex).vData(BlkPtr + t) = (((g_m_Block(p, q) + 16384&) \ 32768) * _
                  '                                                        FScale(t) + 131072) \ 262144
                  '                                                      Keep 1 decimal bit
                  piaSET BYVAL g_Comp(CompIndex).ptrData, BlkPtr + t, (((g_m_Block(p, q) + 16384&) \ 32768) * _
                  FScale(t) + 131072) \ 262144
                  '  2* true scale          4*scale
              NEXT p
            NEXT q

            '--- BlkPtr = BlkPtr + 64
            ! mov eax, BlkPtr
            ! add eax, 64
            ! mov BlkPtr, eax

            '--- i = j + 24
            ! mov eax, j
            ! add eax, 24
            ! mov i, eax


            ' Start Next Block this Row
            '--- k = k - 24
            ! mov eax, k
            ! sub eax, 24
            ! mov k, eax

        LOOP UNTIL k = 0

        '--- i = i + BlkModS
        ! mov ecx, BlkMods
        ! mov eax, i
        ! add eax, ecx
        ! mov i, eax

      LOOP
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error2",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Image quality adjustment
'------------------------------------------------------------------------------------
FUNCTION PBJPEG_SetQuality(BYVAL vQuality AS LONG) AS LONG
  TRY
      ' The JPEG compression standard does not have a formal definition for image Quality.
      ' This implementation defines Quality as an integer value between 1 and 100, and
      ' generates quantization tables based on the value given.
      '
      ' Quality < 50  -  Poor image quality with high compression
      ' Quality = 75  -  Good quality pictures for displaying on a monitor or web page ...
      '                  Typical for general use
      ' Quality = 92  -  High quality with non-optimal compression ... Appropriate for printing ...
      '                  Typical digital camera "max quality" setting
      ' Quality > 95  -  Wasteful ... very poor compression with little image quality improvement.
      '                  Use 24-bit BMP TrueColor if you need quality this high.
      LOCAL i          AS LONG
      LOCAL qvalue      AS LONG
      LOCAL maxvalue    AS LONG
      LOCAL scalefactor AS LONG

      IF g_PP = 12 THEN
        maxvalue = 65535
      ELSE
        maxvalue = 255
      END IF

      IF vQuality > 0 AND vQuality <= 100 THEN
        g_m_Quality = vQuality
      ELSE
        g_m_Quality = 75    ' Else default 75
      END IF

      IF (g_m_Quality < 50) THEN
        IF g_m_Quality = 0 THEN
            scalefactor = 5000
        ELSE
            scalefactor = 5000 \ g_m_Quality
        END IF
      ELSE
        IF g_m_Quality >= 100 THEN
            scalefactor = 0
        ELSE
            scalefactor = 200 - g_m_Quality * 2
        END IF
      END IF

      FOR i = 0 TO 63
        qvalue = (g_QLumin(i) * scalefactor + 50) \ 100

        IF qvalue <= 0 THEN
            qvalue = 1
        ELSEIF qvalue > maxvalue THEN
            qvalue = maxvalue
        END IF

        g_QTable(0).Qk(i) = qvalue
      NEXT i

      FOR i = 0 TO 63
        qvalue = (g_QChrom(i) * scalefactor + 50) \ 100

        IF qvalue <= 0 THEN
            qvalue = 1
        ELSEIF qvalue > maxvalue THEN
            qvalue = maxvalue
        END IF

        g_QTable(1).Qk(i) = qvalue
      NEXT i

      ExpandDQT 0
      ExpandDQT 1
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error3",,ERROR$(ERR)
  END TRY
END FUNCTION

'------------------------------------------------------------------------------------

SUB ExpandDQT(BYVAL Tqi AS LONG)
  TRY
      LOCAL i        AS LONG
      LOCAL j        AS LONG
      LOCAL k        AS LONG
      LOCAL maxvalue AS LONG

      IF g_PP = 12 THEN
        maxvalue = 65535
      ELSE
        maxvalue = 255
      END IF

      FOR i = 0 TO 7
        FOR j = 0 TO 7
            k = g_ZigZag(i, j)

            IF g_QTable(Tqi).Qk(k) < 1 OR g_QTable(Tqi).Qk(k) > maxvalue THEN
              IF %NOMSG = 0 THEN MSGBOX "Bad Quantization Table",,"JPEG encode error"
              ' Bad Quantization Table
            END IF

            g_QTable(Tqi).FScale(k) = (g_EIDCTScale(i, j) + g_QTable(Tqi).Qk(k) \ 2) \ _
            g_QTable(Tqi).Qk(k) ' g_QTable(Tqi).Qk(k)=1..maxval
            ' FScale always 0..65536/131072
        NEXT j
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error4",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Huffman table generation
'------------------------------------------------------------------------------------
SUB BuildHuffman(BYREF TheHuff AS HUFFMANTABLE, BYREF Freq() AS LONG)
  TRY
      ' Generate optimized values for BITS and HUFFVAL in a HUFFMANTABLE
      ' based on symbol frequency counts. Freq must be dimensioned Freq(0-256)
      ' and contain counts of symbols 0-255. Freq is destroyed in this procedure.
      LOCAL i            AS LONG
      LOCAL j            AS LONG
      LOCAL k            AS LONG
      LOCAL n            AS LONG
      LOCAL w            AS LONG
      LOCAL v1            AS LONG
      LOCAL v2            AS LONG
      DIM  others(256)  AS LONG
      DIM  codesize(256) AS LONG
      DIM  vBits(256)    AS LONG
      LOCAL swp          AS LONG
      LOCAL swp2          AS LONG

      ' Initialization
      FOR i = 0 TO 256  ' Initialize others to -1, (this value terminates chain of indicies)
        others(i) = -1
      NEXT i

      Freq(256) = 1    ' Add dummy symbol to guarantee no code will be all '1' bits

      ' Generate codesize() [find huffman code sizes]
      DO                  ' Do loop for (#non-zero-frequencies - 1) times
        v1 = -1                            ' Find highest v1 for least value of freq(v1)>0
        v2 = -1                            ' Find highest v2 for next least value of freq(v2)>0
        swp = 2147483647                  ' Max Long variable
        swp2 = 2147483647

        FOR i = 0 TO 256
            w = Freq(i)

            IF w <> 0 THEN
              IF w <= swp2 THEN
                  IF w <= swp THEN
                    swp2 = swp
                    v2 = v1
                    swp = w
                    v1 = i
                  ELSE
                    swp2 = w
                    v2 = i
                  END IF
              END IF
            END IF
        NEXT i

        IF v2 = -1 THEN
            Freq(v1) = 0                  ' All elements in freq are now set to zero
            EXIT DO                        ' Done
        END IF

        Freq(v1) = Freq(v1) + Freq(v2)    ' Merge the two branches
        Freq(v2) = 0
        codesize(v1) = codesize(v1) + 1    ' Increment all codesizes in v1's branch

        DO WHILE others(v1) >= 0
            v1 = others(v1)
            codesize(v1) = codesize(v1) + 1
        LOOP

        others(v1) = v2                    ' Chain v2 onto v1's branch
        codesize(v2) = codesize(v2) + 1    ' Increment all codesizes in v2's branch

        DO WHILE others(v2) >= 0
            v2 = others(v2)
            codesize(v2) = codesize(v2) + 1
        LOOP
      LOOP

      ' Count BITS [find the number of codes of each size]
      n = 0

      FOR i = 0 TO 256
        w = codesize(i)

        IF w <> 0 THEN
            vBits(w) = vBits(w) + 1
            IF n < w THEN n = w    ' Keep track of largest codesize
        END IF
      NEXT i

      ' Adjust BITS [limit code lengths to 16 bits]
      i = n

      DO WHILE i > 16
        DO WHILE vBits(i) > 0
            FOR j = i - 2 TO 1 STEP -1          ' Since symbols are paired for the longest Huffman
              IF vBits(j) > 0 THEN EXIT FOR  ' code, the symbols are removed from this length
            NEXT j                              ' category two at a time.  The prefix for the pair
            ' (which is one bit shorter) is allocated to one
            vBits(i) = vBits(i) - 2            ' of the pair;  then, (skipping the BITS entry for
            vBits(i - 1) = vBits(i - 1) + 1    ' that prefix length) a code word from the next
            vBits(j + 1) = vBits(j + 1) + 2    ' shortest non-zero BITS entry is converted into
            vBits(j) = vBits(j) - 1            ' a prefix for two code words one bit longer.
        LOOP

        i = i - 1
      LOOP

      vBits(i) = vBits(i) - 1                    ' Remove dummy symbol code from the code length count

      ' Copy BITS and HUFFVAL to the HUFFMANTABLE [HUFFVAL sorted by code length, then by value]
      FOR i = 1 TO 16
        TheHuff.vBits(i - 1) = vBits(i)
      NEXT i

      k = 0

      FOR i = 1 TO n
        FOR j = 0 TO 255
            IF codesize(j) = i THEN
              TheHuff.HuffVal(k) = j
              k = k + 1
            END IF
        NEXT j
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error5",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB ExpandHuffman(BYREF TheHuff AS HUFFMANTABLE, BYVAL MaxSymbol AS LONG)
  TRY
      ' Given a HUFFMANTABLE with valid BITS and HUFFVAL, generate tables for
      ' EHUFCO, EHUFSI, MAXCODE, and MINCODE so the table may be used for compression
      ' and/or decompression.  In JPEG, MaxSymbol is 255 for an AC Huffman Table.  For
      ' DC Tables, MaxSymbol is 11 for PP=8 bit precission, or 15 for PP=12 bit precission.
      LOCAL i      AS LONG  ' Index for BITS
      LOCAL j      AS LONG  ' Index for HUFFVAL
      LOCAL k      AS LONG  ' Index for last HUFFVAL of length (i+1)
      LOCAL si    AS LONG  ' Huffman code size  ( =2^i )
      LOCAL CODE  AS LONG  ' Huffman code
      LOCAL symbol AS LONG  ' Huffman symbol

      FOR i = 0 TO 255
        TheHuff.EHufLN(i) = 0
        TheHuff.EHufSI(i) = 0      ' Clear existing values so we can
        TheHuff.EHufCO(i) = -1    ' Check for duplicate huffman symbols
      NEXT i

      j = 0
      si = 1
      CODE = 0

      FOR i = 0 TO 15
        k = j + TheHuff.vBits(i)

        IF k > 256 THEN
            IF %NOMSG = 0 THEN MSGBOX "Bad Huffman Table (More than 256 symbols)",,"JPEG encode error"
            ' Bad Huffman Table
            ' (More than 256 symbols)
        END IF

        IF j = k THEN                                      ' No codes of length i+1
            TheHuff.MinCode(i) = j - CODE
            TheHuff.MaxCode(i) = -1
        ELSE
            TheHuff.MinCode(i) = j - CODE

            DO WHILE j < k
              symbol = TheHuff.HuffVal(j)                ' Read symbol, make sure it's valid

              IF symbol > MaxSymbol THEN
                  IF %NOMSG = 0 THEN MSGBOX "Bad Huffman Table (Invalid symbol)",,"JPEG encode error"
                  ' Bad Huffman Table
                  ' (Invalid symbol)
              END IF
              IF TheHuff.EHufCO(symbol) >= 0 THEN
                  IF %NOMSG = 0 THEN MSGBOX "Bad Huffman Table (Duplicate symbol)",,"JPEG encode error"
                  ' Bad Huffman Table
                  ' (Duplicate symbol)
              END IF

              TheHuff.EHufLN(symbol) = i
              TheHuff.EHufSI(symbol) = si                ' Assign code for symbol
              TheHuff.EHufCO(symbol) = CODE
              CODE = CODE + 1
              j = j + 1
            LOOP

            TheHuff.MaxCode(i) = CODE - 1
        END IF

        si = si + si

        IF CODE >= si THEN
            IF %NOMSG = 0 THEN MSGBOX "Bad Huffman Table (Code does not fit into available bits)",,"JPEG encode error"
            ' Bad Huffman Table
            ' (Code does not fit into available bits)
        END IF

        CODE = CODE + CODE
      NEXT i

      IF j = 0 THEN
        IF %NOMSG = 0 THEN MSGBOX "Bad Huffman Table (No huffman symbols???)",,"JPEG encode error"
        ' Bad Huffman Table
        ' (No huffman symbols???)
      END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error6",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

FUNCTION CodeLength(BYREF TheHuff AS HUFFMANTABLE, BYREF Freq() AS LONG) AS LONG
  TRY
      ' Estimate a worstcase code length for an entire table
      LOCAL j  AS LONG
      LOCAL k  AS LONG
      LOCAL sum AS LONG

      sum = 0

      FOR j = 0 TO 255
        IF Freq(j) <> 0 THEN
            k = TheHuff.EHufLN(j) + nBits(j)    ' Length of size category + value, in bits
            sum = sum + Freq(j) * k            ' Sum all occurances of this coefficient, in bits
        END IF
      NEXT j

      FUNCTION = (sum + 7) \ 8
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error7",,ERROR$(ERR)
  END TRY
END FUNCTION


'------------------------------------------------------------------------------------
' Entropy coding
'------------------------------------------------------------------------------------
SUB WriteBitsBegin()
  g_m_Chr = 0
  g_m_Bit = 128
END SUB

'------------------------------------------------------------------------------------

SUB WriteBitsEnd()
  IF g_m_Bit <> 128 THEN WriteBits g_m_Bit, -1
END SUB

'------------------------------------------------------------------------------------

SUB WriteBits(BYVAL mask AS LONG, BYVAL CODE AS LONG)
  TRY
      DO WHILE mask > 0
        IF (CODE AND mask) <> 0 THEN g_m_Chr = (g_m_Chr OR g_m_Bit)

        IF g_m_Bit = 1 THEN            ' We completed a byte ...
            g_m_Data(g_m_Ptr) = g_m_Chr    ' Add it to the stream

            IF g_m_Chr = 255 THEN      ' Pad a zero byte and advance pointer
              g_m_Ptr = g_m_Ptr + 1    ' Advance pointer
              g_m_Data(g_m_Ptr) = 0    ' Insert zero
            END IF

            g_m_Ptr = g_m_Ptr + 1        ' Advance pointer
            g_m_Chr = 0                ' Clear byte buffer and reset bit index
            g_m_Bit = 128
        ELSE                          ' Increment to next bit position to write
            g_m_Bit = g_m_Bit \ 2
        END IF

        mask = mask \ 2
      LOOP
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error8",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB EncodeCoefficients(BYREF vData() AS INTEGER, BYREF p AS LONG, BYREF Pred AS LONG, _
  BYVAL Td AS LONG, BYVAL Ta AS LONG)
  TRY
      ' Use Huffman tables to compress a block of 64 quantized DCT coefficients to the local
      ' g_m_Data() byte array.  The coefficients are input in the data() array starting at index p.
      ' Pred is the predictor for the DC coefficient.  Td and Ta are indexes to the local DC and AC
      ' Huffman Tables to use.
      LOCAL r    AS LONG
      LOCAL rs  AS LONG
      LOCAL CODE AS LONG
      LOCAL p2  AS LONG

      p2 = p + 64

      CODE = vData(p) - Pred
      Pred = vData(p)
      p = p + 1

      IF CODE = 0 THEN
        WriteBits g_HuffDC(Td).EHufSI(0), g_HuffDC(Td).EHufCO(0)      ' Append symbol for size category
      ELSE
        rs = nBits(CODE)
        IF CODE < 0 THEN CODE = CODE - 1
        WriteBits g_HuffDC(Td).EHufSI(rs), g_HuffDC(Td).EHufCO(rs)    ' Append symbol for size category
        WriteBits g_TwoR(rs), CODE                                  ' Append diff
      END IF

      r = 0

      DO
        CODE = vData(p)

        IF CODE = 0 THEN
            r = r + 16
        ELSE
            DO WHILE r > 240
              WriteBits g_HuffAC(Ta).EHufSI(240), g_HuffAC(Ta).EHufCO(240)  ' Append RUN16 (a run of 16 zeros)
              r = r - 256
            LOOP

            rs = nBits(CODE)
            IF CODE < 0 THEN CODE = CODE - 1
            WriteBits g_HuffAC(Ta).EHufSI(r + rs), g_HuffAC(Ta).EHufCO(r + rs) ' Append run length, size category
            WriteBits g_TwoR(rs), CODE                                      ' Append AC value
            r = 0
        END IF

        p = p + 1
      LOOP WHILE p < p2  ' Should be equal on exit

      IF r <> 0 THEN WriteBits g_HuffAC(Ta).EHufSI(0), g_HuffAC(Ta).EHufCO(0)    ' Append EOB (end of block)
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error9",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Collecting statistics
'------------------------------------------------------------------------------------
' These procedures collect statistics of run-length and size categories of DCT coefficients
' so optimized Huffman tables can be generated to compress them.
SUB CollectStatisticsAC(BYREF vData() AS INTEGER, BYREF FreqAC() AS LONG)
  TRY
      LOCAL CODE AS LONG
      LOCAL n    AS LONG    ' Number of coefficients in data()
      LOCAL p    AS LONG    ' Index for current data() coefficient
      LOCAL p2  AS LONG
      LOCAL r    AS LONG    ' Run length of zeros
      LOCAL rs  AS LONG    ' Run-length/Size-category Symbol

      n = UBOUND(vData) + 1 ' +1 not necessary
      p = 0

      DO WHILE p < n
        p = p + 1
        p2 = p + 63
        r = 0

        DO WHILE p < p2  ' Should be equal on exit
            CODE = vData(p)                            ' Is always <=|255| ??

            IF CODE = 0 THEN
              r = r + 16                            ' 16*count of zeroes
            ELSE
              DO WHILE r > 240
                  FreqAC(240) = FreqAC(240) + 1    ' Runs of 16 zeros
                  r = r - 256
              LOOP

              rs = r + nBits(CODE)                  ' Number of bits for this code
              FreqAC(rs) = FreqAC(rs) + 1          ' Freq of 16*nzeroes + nbits(code)+1
              r = 0
            END IF

            p = p + 1
        LOOP

        IF r <> 0 THEN FreqAC(0) = FreqAC(0) + 1      ' EOB value
      LOOP
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error10",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB CollectStatisticsDCNonInterleaved(BYREF vData() AS INTEGER, BYREF FreqDC() AS LONG)
  TRY
      LOCAL Pred AS LONG    ' DC Predictor
      LOCAL n    AS LONG    ' Number of coefficients in data()
      LOCAL p    AS LONG    ' Index for current data() coefficient
      LOCAL rs  AS LONG    ' Size category for Diff

      n = UBOUND(vData) + 1
      p = 0
      Pred = 0

      DO WHILE p < n
        rs = nBits(vData(p) - Pred)  ' Number of bits in delta from last value
        Pred = vData(p)
        FreqDC(rs) = FreqDC(rs) + 1  ' Frequencies of bitsizes
        p = p + 64
      LOOP
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error11",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB CollectStatisticsDCInterleaved(BYREF vData() AS INTEGER, BYREF FreqDC() AS LONG, _
  BYVAL vHi AS LONG, BYVAL Vi AS LONG)
  TRY
      LOCAL p()  AS LONG    ' Index to .data in component f for scanline g
      LOCAL f    AS LONG    ' Index counter (component)
      LOCAL g    AS LONG    ' Index counter (sampling factor, vertical)
      LOCAL h    AS LONG    ' Index counter (sampling factor, horizontal)
      LOCAL i    AS LONG    ' Index counter (MCU horizontal)
      LOCAL j    AS LONG    ' Index counter (MCU vertical)
      LOCAL n    AS LONG    ' Number of coefficients in data()
      LOCAL s    AS LONG    ' Size category for Diff
      LOCAL Pred AS LONG    ' DC Predictor
      LOCAL pLF  AS LONG    ' Line Feed for p in data
      LOCAL MCUr AS LONG    ' Number of complete 8X8 blocks in rightmost MCU
      LOCAL MCUx AS LONG    ' Number of MCUs per scanline
      LOCAL MCUy AS LONG    ' Number of MCU scanlines

      n = UBOUND(vData) + 1
      REDIM p(Vi - 1)

      MCUx = (g_XX + 8 * g_HMax - 1) \ (8 * g_HMax)
      MCUy = ( g_YY + 8 * g_VMax - 1) \ (8 * g_VMax)

      h = (-INT(-g_XX * vHi / g_HMax) + 7) \ 8 ' Width of scanline in data (MCUs)

      FOR g = 0 TO Vi - 1                  ' Initialize .data pointers
        p(g) = 64 * h * g
      NEXT g

      pLF = 64 * h * (Vi - 1)              ' Initialize .data pointer advancer

      MCUr = (h MOD vHi)                  ' Number of complete 8X8 Blocks in rightmost MCU
      IF MCUr = 0 THEN MCUr = vHi

      FOR j = 1 TO MCUy - 1
        ' MCUs across a scanline
        FOR i = 1 TO MCUx - 1
            FOR g = 1 TO Vi
              FOR h = 1 TO vHi
                  s = nBits(vData(p(g - 1)) - Pred)  ' Number of bits in diff
                  Pred = vData(p(g - 1))
                  FreqDC(s) = FreqDC(s) + 1
                  p(g - 1) = p(g - 1) + 64
              NEXT h
            NEXT g
        NEXT i

        ' Rightmost MCU
        FOR g = 1 TO Vi
            FOR h = 1 TO vHi
              IF h > MCUr THEN                      ' Pad with dummy block
                  s = 0
              ELSE
                  s = nBits(vData(p(g - 1)) - Pred)  ' Number of bits in diff
                  Pred = vData(p(g - 1))
                  p(g - 1) = p(g - 1) + 64
              END IF

              FreqDC(s) = FreqDC(s) + 1
            NEXT h
        NEXT g

        ' Advance data pointers
        FOR g = 0 TO Vi - 1
            p(g) = p(g) + pLF
        NEXT g
      NEXT j

      ' Bottommost MCU Scanline
      FOR i = 1 TO MCUx
        FOR g = 1 TO Vi
            FOR h = 1 TO vHi
              IF p(g - 1) >= n OR (i = MCUx AND h > MCUr) THEN ' Pad with dummy block
                  s = 0
              ELSE
                  s = nBits(vData(p(g - 1)) - Pred)            ' Number of bits in Diff
                  Pred = vData(p(g - 1))
                  p(g - 1) = p(g - 1) + 64
              END IF

              FreqDC(s) = FreqDC(s) + 1
            NEXT h
        NEXT g
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error12",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Image sampling
'------------------------------------------------------------------------------------
' This class always samples and compresses pictures in YCbCr colorspace.  The first component, Y,
' represents the Luminance of the pixels.  This is "how bright" a pixel is.  The Cb and Cr
' components are Chrominance, which is a measure of how far from neutral-white (toward a color)
' a pixel is.  The human visual sensory system can discriminate Luminance differences about
' twice as well as it can discriminate Chrominance differences.
'
' Virtually all JPEG files are in YCbCr colorspace.  Other JPEG compliant colorspaces exist, but
' they are used in specialty equipment.  For example, people in the astronomy or medical fields
' choose colorspaces that best record the information they are interested in, and don't care about
' how pretty the picture looks to a person when displayed on a computer monitor.
' [Apple/Machintosh sometimes uses a four component colorspace, but that colorspace is rare and
' not widely supported]
'
' Sampling frequencies define how often each component is sampled.  Higher frequencies store more
' information, while lower frequencies store less.  Typically, sampling frequencies are set at
' 2,2, 1,1, 1,1.  This corresponds to the human visual sensory system.  The first component,
' Luminance, is sampled twice as much because our eyes notice differences in Luminance quite easily.
' The two Chrominance components are sampled half as much as because our eyes can't distinguish
' the difference in color changes as well.  One Luminance value is sampled for every pixel, and
' one Chrominance value is sampled for each 2X2 block of pixels.
'
' Digital cameras typically record at sampling frequencies of 1,1, 1,1, 1,1.  This samples every
' pixel for all three components.  The quality of the picture is a little better when viewed by
' a person, but the compression benefits drop significantly.  If the picture to be compressed
' is from a Scanner or Digital camera, and you plan on printing it in the future, and storage
' space is not a problem, then sampling at these frequencies makes sense.  Otherwise, if you only
' plan on using the picture to display on a monitor or a web page, [2,2, 1,1, 1,1] makes the
' most sense.
'
' The JPEG standard specifies that sampling frequencies may range from 1-4 for each component
' in both directions.  However, if any component has a sampling frequency of '3', and another
' component has a coresponding sampling frequency of '2' or '4', the downsampling process
' will map fractional pixels to sample values.  This is leagal in the JPEG standard, and this
' class will compress fractional pixel samplings, but this is not widely supported.  It is
' highly recommended to AVOID SAMPLING FACTORS OF 3 for maximum compatability with JPEG decoders.
'
' Some JPEG encoders avoid the fractional pixel problem by only allowing the end user to pick
' a "sub-sampling" value.  In such "Sub Sampling" schemes, all Chrominance frequencies are set
' to one, and the (one or two) sub-sampling value(s) specify Luminance frequencies.
'
' There should *never* be an error raised if you are using this class correctly.  It should
' not be possible for the end user to specify illegal sampling frequency values!
' [For tinkerers - If you delete the error raising code and specify illegal sampling
' frequencies, this class will procede to create a non-JPEG compliant file with the values
' specified]
SUB PBJPEG_SetSamplingFrequencies(BYVAL H1 AS LONG, BYVAL v1 AS LONG, _
  BYVAL H2 AS LONG, BYVAL v2 AS LONG, _
  BYVAL H3 AS LONG, BYVAL v3 AS LONG)
  LOCAL i AS LONG

  TRY

      IF H1 < 1 OR H1 > 4 THEN
        IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err1"
        ' Invalid Sampling Value
      END IF
      IF v1 < 1 OR v1 > 4 THEN
        ' Invalid Sampling Value
        IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err2"
      END IF

      IF (H2 OR H3 OR v2 OR v3) = 0 THEN  ' If H2,H3,V2,V3 are all zero ...
        g_Nf = 1          ' Luminance only.
        REDIM g_Comp(0)
        g_Comp(0).vHi = 1 ' Set up for sampling Greyscale
        g_Comp(0).Vi = 1  ' (Black and White picture)
        g_Comp(0).ptrData = VARPTR(g_CompData0())
      ELSE
        IF H2 < 1 OR H2 > 4 THEN
            ' Invalid Sampling Value
            IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err3"
        END IF
        IF H3 < 1 OR H3 > 4 THEN
            ' Invalid Sampling Value
            IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err4"
        END IF
        IF v2 < 1 OR v2 > 4 THEN
            ' Invalid Sampling Value
            IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err5"
        END IF
        IF v3 < 1 OR v3 > 4 THEN
            ' Invalid Sampling Value
            IF %NOMSG = 0 THEN MSGBOX "Invalid Sampling Value",,"Err6"
        END IF

        g_Nf = 3          ' YCbCr
        REDIM g_Comp(2)

        g_Comp(0).vHi = H1
        g_Comp(0).Vi = v1
        g_Comp(0).Tqi = 0
        g_Comp(0).ptrData = VARPTR(g_CompData0())
        g_Comp(1).vHi = H2
        g_Comp(1).Vi = v2
        g_Comp(1).Tqi = 1
        g_Comp(1).ptrData = VARPTR(g_CompData1())
        g_Comp(2).vHi = H3
        g_Comp(2).Vi = v3
        g_Comp(2).Tqi = 1
        g_Comp(2).ptrData = VARPTR(g_CompData2())
      END IF

      g_HMax = -1
      g_VMax = -1

      FOR i = 0 TO g_Nf - 1  ' Determine max h, v sampling factors
        IF g_HMax < g_Comp(i).vHi THEN g_HMax = g_Comp(i).vHi
        IF g_VMax < g_Comp(i).Vi THEN g_VMax = g_Comp(i).Vi
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error13",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

FUNCTION PBJPEG_SampleHDC(BYVAL lHDC AS LONG, _
  BYVAL lWidth AS LONG, _
  BYVAL lHeight AS LONG, _
  BYVAL lSrcLeft AS LONG, _
  BYVAL lSrcTop AS LONG) AS LONG
  ' Given a valid hDC and dimensions, generate component samplings of an Image.
  ' A DIBSection is created to hold Sample(s) of the Image, from which the Image is
  ' decomposed into YCbCr components.
  ' Returns: 0 = Success
  '          1 = API error while generating a DIBSection
  TRY
      LOCAL BI      AS BITMAPINFO ' Type containing the Bitmap information
      LOCAL hDIb    AS LONG      ' Handle to the DIBSection
      LOCAL hBmpOld  AS LONG      ' Handle to the old bitmap in the DC, for clear up
      LOCAL hDC      AS LONG      ' Handle to the Device context holding the DIBSection
      LOCAL pBits    AS DWORD PTR  ' Address of memory pointing to the DIBSection's bits

      LOCAL f        AS LONG      ' Index counter for components
      LOCAL g        AS LONG
      LOCAL vPixel() AS BYTE      ' Byte array containing pixel data
      LOCAL lRowMod  AS LONG      ' Width of A Row in Bytes

      LOCAL xs      AS LONG      ' Sample width
      LOCAL ys      AS LONG      ' Sample height
      LOCAL xs8      AS LONG      ' Sample width  (padded to 8 pixel barrier)
      LOCAL ys8      AS LONG      ' Sample height (padded to 8 pixel barrier)
      LOCAL xsLast  AS LONG      ' Sample width  (for previous component)
      LOCAL ysLast  AS LONG      ' Sample height (for previous component)
      DIM FScale(63) AS LONG

      g_PP = 8
      g_XX = lWidth                ' The to be sampled image block
      g_YY = lHeight              ' The to be sampled image block

      ' Create a DIBSection to store Sampling(s) of the Image
      hDC = CREATECOMPATIBLEDC(0)

      IF hDC = 0 THEN
        FUNCTION = 1        ' CreateCompatibleDC() API Failure
        IF %NOMSG = 0 THEN MSGBOX "CreateCompatibleDC() API Failure",,"JPEG encode error"
      ELSE


        'BI.bmiHeader.biSizeImage = ((BI.bmiHeader.biWidth * 3 + 3) And &HFFFFFFFC) * BI.bmiHeader.biHeight '4 byte barrier
        'hDIb = CreateDIBSection2(hDC, BI, DIB_RGB_COLORS, lPtr, 0, 0)

        lWidth = (g_XX + 7) AND &HFFFFFFF8    ' 8 byte barrier for 8X8 data units
        lHeight = (g_YY + 7) AND &HFFFFFFF8
        'MsgBox "width=" + Str$(lWidth) + "height=" + Str$(lHeight)
        lRowMod = ((lWidth * 3 + 3) AND &HFFFFFFFC)'(((lWidth * 24&) + 31&) And Not 31&) \ 8&
        BI.bmiHeader.biSize = LEN(BI.bmiHeader)
        BI.bmiHeader.biWidth = lWidth
        BI.bmiHeader.biHeight = -lHeight    ' Top to bottom please
        BI.bmiHeader.biPlanes = 1
        BI.bmiHeader.biBitCount = 24
        BI.bmiHeader.biCompression = %BI_RGB
        BI.bmiHeader.biSizeImage = (lRowMod * lHeight)  ' 4 byte barrier

        hDIb = CREATEDIBSECTION(hDC, BI, %DIB_RGB_COLORS, BYVAL VARPTR(pBits), 0, 0)

        IF hDIb = 0 THEN
            FUNCTION = 1    ' CreateDIBSection() API Failure
            IF %NOMSG = 0 THEN MSGBOX "CreateDIBSection() API Failure",,"JPEG encode error"
        ELSE
            GdiFlush

            hBmpOld = SELECTOBJECT(hDC, hDIb)          ' Select DIBSection into DC

            IF SETSTRETCHBLTMODE(hDC, %HALFTONE) = 0 THEN SETSTRETCHBLTMODE hDC, %COLORONCOLOR

            FOR f = 0 TO g_Nf - 1
              g_Comp(f).Ci = f + 1    ' Assign an ID to this component

              xs = -INT(-g_XX * g_Comp(f).vHi / g_HMax)    ' Determine Sample dimensions
              ys = -INT(-g_YY * g_Comp(f).Vi / g_VMax)
              xs8 = ((xs + 7) AND &HFFFFFFF8)        ' Sample dimensions with 8X8 barrier
              ys8 = ((ys + 7) AND &HFFFFFFF8)

              IF xs8 <> xsLast OR ys8 <> ysLast THEN ' We need to Sample the Image
                  IF xs = g_XX AND ys = g_YY THEN        ' Just copy the image to our DIBSection
                    BITBLT hDC, 0, 0, xs, ys, lHDC, lSrcLeft, lSrcTop, %SRCCOPY
                  ELSE                              ' Resample/Resize the Image
                    STRETCHBLT hDC, 0, 0, xs, ys, lHDC, lSrcLeft, lSrcTop, lWidth, lHeight, %SRCCOPY
                  END IF

                  IF xs < xs8 THEN                  ' Pad right of Sample to 8 block barrier
                    STRETCHBLT hDC, xs + 1, 0, xs8 - xs, ys, hDC, xs, 0, 1, ys, %SRCCOPY
                  END IF

                  IF ys < ys8 THEN                  ' Pad bottom of Sample to 8 block barrier
                    STRETCHBLT hDC, 0, ys + 1, xs8, ys8 - ys, hDC, 0, ys, xs8, 1, %SRCCOPY
                  END IF
              END IF

              ' Copy the DIB section's pixels to the vPixel array
              ' This is so dirty it hurts
              REDIM vPixel((lRowMod * lHeight)-1)

              'LogThis "Starting vPixel mod 4 = " + Str$((UBound(vPixel)+1) Mod 4)

              FOR g = 0 TO UBOUND(vPixel)
                  'If (g Mod 100000) = 0 Or (g > 2359196) Then LogThis "on vPixel " + Str$(g)
                  vPixel(g) = PEEK(pBits + g)      '<------  had GPF on large images on second encoding
              NEXT g

              'MoveMemory (ByVal VarPtr(vPixel(0)), _ '        // address Of Copy destination
              '            ByVal pBits,            _ ' *Source,        // address Of block To Copy
              '            ByVal UBound(vPixel) + 1)  ' Length        // Size, In bytes, Of block To Copy


              'LogThis "UBound(vPixel)=" + Str$(UBound(vPixel))

              FOR g = 0 TO UBOUND(FScale)
                  FScale(g) = g_QTable(g_Comp(f).Tqi).FScale(g)
              NEXT g

              ' Read 8X8 blocks of pixels, convert from RGB->YCbCr colorspace, FDCT and Quantize
              ' the data, store the results in .Data() of this component
              CALL TransformAndQuantize(vPixel(), f, FScale(), xs8, ys8, lRowMod)
              'CALL TransformAndQuantize(vPixel(), f, g_QTable(g_Comp(f).Tqi).FScale(), xs8, ys8, lRowMod)

              xsLast = xs
              ysLast = ys
            NEXT f

            SELECTOBJECT hDC, hBmpOld ' Select CompatibleDC  (unselect DIBSection)
            DELETEOBJECT hDIb        ' Delete DIBSection
        END IF

        DELETEDC hDC              ' Delete CompatibleDC
      END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error14",,ERROR$(ERR)
  END TRY
END FUNCTION

'------------------------------------------------------------------------------------

FUNCTION PBJPEG_AddComment(BYVAL vComment AS STRING) AS LONG
  ' Assigning a value to this property will add the text Comment to the JPEG file.
  IF LEN(vComment) > 65535 THEN
      IF %NOMSG = 0 THEN MSGBOX "Illegal Comment Length",,"JPEG encode error"
      ' Illegal Comment Length
  ELSE
      g_m_Comment = vComment
  END IF

END FUNCTION


'------------------------------------------------------------------------------------
' Emiting markers
'------------------------------------------------------------------------------------
SUB InsertJFIF()
  TRY
      IF g_m_Ptr + 17 > UBOUND(g_m_Data) THEN
        IF %NOMSG = 0 THEN MSGBOX "Copymemory will write past bounds of g_m_Data()",,"JPEG encode error"
        ' Copymemory will write past bounds of g_m_Data()
      ELSE
        POKE VARPTR(g_m_Data(g_m_Ptr + 0)), &H1000E0FF    ' APP0 Marker, Length(APP0)=16
        POKE VARPTR(g_m_Data(g_m_Ptr + 4)), &H4649464A    ' "JFIF"
        POKE VARPTR(g_m_Data(g_m_Ptr + 8)), &H10100      ' "/0", Version Major=1, Version Minor=1
        ' Units=0  [0=pixel, 1=dpi, 2=dots/cm]
        POKE VARPTR(g_m_Data(g_m_Ptr + 12)), &H1000100    ' Horizontal pixel density = 1 (dot per pixel)
        ' Vertical pixel density = 1 (dot per pixel)
        POKE VARPTR(g_m_Data(g_m_Ptr + 16)), &H0          ' Thumbnail horizontal pixel count = 0
        g_m_Ptr = g_m_Ptr + 18                                  ' Thumbnail vertical pixel count = 0
      END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error15",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB InsertSOF(BYVAL SOFMarker AS LONG)
  TRY
      LOCAL i  AS LONG  ' Insert a Start Of Frame marker segment
      LOCAL Lx AS LONG  ' PP, YY, XX, Nf, and Ci,Hi,Vi,Tqi, must already be set

      Lx = 8 + 3 * g_Nf
      g_m_Data(g_m_Ptr) = 255                    ' SOF
      g_m_Data(g_m_Ptr + 1) = SOFMarker AND 255
      g_m_Data(g_m_Ptr + 2) = Lx \ 256          ' Frame Header Length
      g_m_Data(g_m_Ptr + 3) = Lx AND 255
      g_m_Data(g_m_Ptr + 4) = g_PP                ' Sample precision [8, 12]
      g_m_Data(g_m_Ptr + 5) = g_YY \ 256          ' Number of Lines
      g_m_Data(g_m_Ptr + 6) = g_YY AND 255
      g_m_Data(g_m_Ptr + 7) = g_XX \ 256          ' Number of samples per line
      g_m_Data(g_m_Ptr + 8) = g_XX AND 255
      g_m_Data(g_m_Ptr + 9) = g_Nf                ' Number of image components in frame
      g_m_Ptr = g_m_Ptr + 10

      FOR i = 0 TO g_Nf - 1                    ' For each component ...
        g_m_Data(g_m_Ptr) = g_Comp(i).Ci        ' Component identifier
        g_m_Data(g_m_Ptr + 1) = g_Comp(i).vHi * 16 OR g_Comp(i).Vi  ' Horizontal/Vertical sampling factors
        g_m_Data(g_m_Ptr + 2) = g_Comp(i).Tqi    ' Quantization table selector
        g_m_Ptr = g_m_Ptr + 3
      NEXT i
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error16",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB InsertCOM(BYVAL TheComment AS STRING)
  TRY
      LOCAL i  AS LONG
      LOCAL Lx AS LONG

      Lx = LEN(TheComment) + 2

      IF Lx > 2 THEN
        g_m_Data(g_m_Ptr) = 255              ' COM marker
        g_m_Data(g_m_Ptr + 1) = %COM
        g_m_Data(g_m_Ptr + 2) = Lx \ 256      ' COM marker segment length
        g_m_Data(g_m_Ptr + 3) = Lx AND 255
        g_m_Ptr = g_m_Ptr + 4

        FOR i = 1 TO LEN(TheComment)      ' Comment text
            g_m_Data(g_m_Ptr) = ASC(MID$(TheComment, i, 1))
            g_m_Ptr = g_m_Ptr + 1
        NEXT i
      END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error17",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB InsertDQT(BYVAL MarkerPos AS LONG, BYVAL Tqi AS LONG)
  TRY
      LOCAL i AS LONG  ' Call with MarkerPos = g_m_Ptr to insert a single table with its own DQT marker
      ' Call multiple times with the same MarkerPos to include
      ' multiple tables under the same DQT marker

      IF g_m_Ptr < MarkerPos + 4 THEN  ' Insert Marker
        g_m_Ptr = MarkerPos + 4
        g_m_Data(g_m_Ptr - 4) = 255
        g_m_Data(g_m_Ptr - 3) = %DQT
      END IF

      FOR i = 0 TO 63
        IF g_QTable(Tqi).Qk(i) > 255 THEN EXIT FOR
      NEXT i

      IF i = 64 THEN                  ' 8 bit precision
        g_m_Data(g_m_Ptr) = Tqi
        g_m_Ptr = g_m_Ptr + 1

        FOR i = 0 TO 63
            g_m_Data(g_m_Ptr) = g_QTable(Tqi).Qk(i)
            g_m_Ptr = g_m_Ptr + 1
        NEXT i
      ELSE                            ' 16 bit precision
        IF g_PP <> 12 THEN
            IF %NOMSG = 0 THEN MSGBOX "Illegal precision in Quantization Table",,"JPEG encode error"
            ' Illegal precision in Quantization Table
        END IF

        g_m_Data(g_m_Ptr) = Tqi OR 16
        g_m_Ptr = g_m_Ptr + 1

        FOR i = 0 TO 63
            g_m_Data(g_m_Ptr) = g_QTable(Tqi).Qk(i) \ 256
            g_m_Data(g_m_Ptr + 1) = g_QTable(Tqi).Qk(i) AND 255
            g_m_Ptr = g_m_Ptr + 2
        NEXT i
      END IF

      g_m_Data(MarkerPos + 2) = (g_m_Ptr - MarkerPos - 2) \ 256&  ' Insert Marker segment length
      g_m_Data(MarkerPos + 3) = (g_m_Ptr - MarkerPos - 2) AND 255&
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error18",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB InsertDHT(BYVAL MarkerPos AS LONG, BYVAL HIndex AS LONG, BYVAL isAC AS LONG)
  TRY
      LOCAL i AS LONG  ' Call with MarkerPos = g_m_Ptr to insert a single table with its own DHT marker
      LOCAL j AS LONG  ' Call multiple times with the same MarkerPos to include
      ' multiple tables under the same DHT marker

      IF g_m_Ptr < MarkerPos + 4 THEN  ' Insert Marker
        g_m_Ptr = MarkerPos + 4
        g_m_Data(g_m_Ptr - 4) = 255
        g_m_Data(g_m_Ptr - 3) = %DHT
      END IF

      IF isAC THEN
        g_m_Data(g_m_Ptr) = HIndex OR 16
        g_m_Ptr = g_m_Ptr + 1
        j = 0

        FOR i = 0 TO 15
            g_m_Data(g_m_Ptr) = g_HuffAC(HIndex).vBits(i)
            g_m_Ptr = g_m_Ptr + 1
            j = j + g_HuffAC(HIndex).vBits(i)
        NEXT i

        FOR i = 0 TO j - 1
            g_m_Data(g_m_Ptr) = g_HuffAC(HIndex).HuffVal(i)
            g_m_Ptr = g_m_Ptr + 1
        NEXT i
      ELSE
        g_m_Data(g_m_Ptr) = HIndex
        g_m_Ptr = g_m_Ptr + 1
        j = 0

        FOR i = 0 TO 15
            g_m_Data(g_m_Ptr) = g_HuffDC(HIndex).vBits(i)
            g_m_Ptr = g_m_Ptr + 1
            j = j + g_HuffDC(HIndex).vBits(i)
        NEXT i

        FOR i = 0 TO j - 1
            g_m_Data(g_m_Ptr) = g_HuffDC(HIndex).HuffVal(i)
            g_m_Ptr = g_m_Ptr + 1
        NEXT i
      END IF

      g_m_Data(MarkerPos + 2) = (g_m_Ptr - MarkerPos - 2) \ 256&  ' Insert Marker segment length
      g_m_Data(MarkerPos + 3) = (g_m_Ptr - MarkerPos - 2) AND 255&
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error19",,ERROR$(ERR)
  END TRY

END SUB

'------------------------------------------------------------------------------------

SUB InsertMarker(BYVAL TheMarker AS LONG)
  g_m_Data(g_m_Ptr) = 255
  g_m_Data(g_m_Ptr + 1) = TheMarker
  g_m_Ptr = g_m_Ptr + 2
END SUB


'------------------------------------------------------------------------------------
' Emiting scans
'------------------------------------------------------------------------------------
SUB InsertSOSNonInterleaved(BYVAL CompIndex AS LONG, BYVAL Td AS LONG, BYVAL Ta AS LONG)
  TRY
      ' Insert an SOS marker and scan data for a non-interleaved Sequential scan.
      LOCAL p    AS LONG    ' Index to .data in component
      LOCAL n    AS LONG
      LOCAL Pred AS LONG    ' Predictor for DC coefficient

      ' Insert SOS Marker Segment
      g_m_Data(g_m_Ptr) = 255                    ' SOS Marker
      g_m_Data(g_m_Ptr + 1) = %SOS
      g_m_Data(g_m_Ptr + 2) = 8 \ 256            ' Marker Segment Length
      g_m_Data(g_m_Ptr + 3) = 8 AND 255
      g_m_Data(g_m_Ptr + 4) = 1                  ' Ns    - Number of components in Scan [1-4]
      g_m_Ptr = g_m_Ptr + 5
      g_m_Data(g_m_Ptr) = g_Comp(CompIndex).Ci    ' Csj    - Component ID
      g_m_Data(g_m_Ptr + 1) = Td * 16 OR Ta      ' Td, Ta - DC, AC entropy coder selector
      g_m_Ptr = g_m_Ptr + 2
      g_m_Data(g_m_Ptr) = 0                      ' Ss    - Start of spectral selection
      g_m_Data(g_m_Ptr + 1) = 63                ' Se    - End of spectral selection
      g_m_Data(g_m_Ptr + 2) = 0                  ' Ah, Al - Successive approximation bit high/low
      g_m_Ptr = g_m_Ptr + 3

      ' Insert non-interleaved sequential entropy coded data
      p = 0
      'n = UBOUND(g_Comp(CompIndex).vData) + 1
      n = piaUBOUND(BYVAL g_Comp(CompIndex).ptrData) + 1
      Pred = 0

      WriteBitsBegin

      DO WHILE p <> n
        EncodeCoefficients BYVAL g_Comp(CompIndex).ptrData, p, Pred, Td, Ta
      LOOP

      WriteBitsEnd
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error20",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

SUB InsertSOSInterleaved(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
  BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG)
  TRY
      ' Insert an SOS marker and scan data for an interleaved Sequential scan.
      LOCAL f        AS LONG      ' Index counter (component)
      LOCAL g        AS LONG      ' Index counter (sampling factor, vertical)
      LOCAL h        AS LONG      ' Index counter (sampling factor, horizontal)
      LOCAL i        AS LONG      ' Index counter (MCU horizontal)
      LOCAL j        AS LONG      ' Index counter (MCU vertical)
      LOCAL Lx        AS LONG      ' Marker Segment Length
      LOCAL Ns        AS LONG      ' Number of components in Scan [1-4]
      LOCAL MCUx      AS LONG      ' Number of MCUs per scanline
      LOCAL MCUy      AS LONG      ' Number of MCU scanlines

      LOCAL p()      AS LONG      ' Index to .data in component f for scanline g
      LOCAL pLF()    AS LONG      ' Line Feed for p in .data for component f
      LOCAL Pred()    AS LONG      ' Predictor for DC coefficient in component f
      LOCAL MCUr()    AS LONG      ' Number of complete 8X8 blocks in rightmost MCU for component f
      DIM  Pad64(63) AS INTEGER  ' 8X8 padding block for completing MCUs

      Ns = SecondIndex - FirstIndex + 1
      Lx = 6 + 2 * Ns

      ' Insert SOS Marker Segment
      g_m_Data(g_m_Ptr) = 255                          ' SOS Marker
      g_m_Data(g_m_Ptr + 1) = %SOS
      g_m_Data(g_m_Ptr + 2) = Lx \ 256                ' Marker Segment Length
      g_m_Data(g_m_Ptr + 3) = Lx AND 255
      g_m_Data(g_m_Ptr + 4) = Ns                      ' Ns    - Number of components in Scan [1-4]
      g_m_Ptr = g_m_Ptr + 5

      FOR i = FirstIndex TO SecondIndex
        g_m_Data(g_m_Ptr) = g_Comp(CompIndex(i)).Ci    ' Csj
        g_m_Data(g_m_Ptr + 1) = Td(i) * 16 OR Ta(i)  ' Td, Ta
        g_m_Ptr = g_m_Ptr + 2
      NEXT i

      g_m_Data(g_m_Ptr) = 0                            ' Ss    - Start of spectral selection
      g_m_Data(g_m_Ptr + 1) = 63                      ' Se    - End of spectral selection
      g_m_Data(g_m_Ptr + 2) = 0                        ' Ah, Al - Successive approximation bit high/low
      g_m_Ptr = g_m_Ptr + 3

      ' Insert interleaved sequential entropy coded data
      REDIM p(FirstIndex TO SecondIndex, g_VMax - 1)
      REDIM Pred(FirstIndex TO SecondIndex)
      REDIM pLF(FirstIndex TO SecondIndex)
      REDIM MCUr(FirstIndex TO SecondIndex)

      MCUx = (g_XX + 8 * g_HMax - 1) \ (8 * g_HMax)
      MCUy = (g_YY + 8 * g_VMax - 1) \ (8 * g_VMax)

      FOR f = FirstIndex TO SecondIndex
        h = (-INT(-g_XX * g_Comp(CompIndex(f)).vHi / g_HMax) + 7) \ 8  ' Width of scanline in .data (MCUs)

        FOR g = 0 TO g_Comp(CompIndex(f)).Vi - 1                  ' Initialize .data pointers
            p(f, g) = 64 * h * g
        NEXT g

        pLF(f) = 64 * h * (g_Comp(CompIndex(f)).Vi - 1)            ' Initialize .data pointer advancer

        MCUr(f) = (h MOD g_Comp(CompIndex(f)).vHi)                ' Number of complete 8X8 Blocks in rightmost MCU
        IF MCUr(f) = 0 THEN MCUr(f) = g_Comp(CompIndex(f)).vHi
      NEXT f

      WriteBitsBegin

      FOR j = 1 TO MCUy - 1
        ' Encode MCUs across a scanline
        FOR i = 1 TO MCUx - 1
            FOR f = FirstIndex TO SecondIndex ' 0 To Ns - 1
              FOR g = 1 TO g_Comp(CompIndex(f)).Vi
                  FOR h = 1 TO g_Comp(CompIndex(f)).vHi
                    EncodeCoefficients BYVAL g_Comp(CompIndex(f)).ptrData, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  NEXT h
              NEXT g
            NEXT f
        NEXT i

        ' Encode Rightmost MCU
        FOR f = FirstIndex TO SecondIndex  ' 0 To Ns - 1
            FOR g = 1 TO g_Comp(CompIndex(f)).Vi
              FOR h = 1 TO g_Comp(CompIndex(f)).vHi
                  IF h > MCUr(f) THEN    ' Pad with dummy block
                    Pad64(0) = Pred(f)
                    EncodeCoefficients Pad64(), 0, Pred(f), Td(f), Ta(f)
                  ELSE
                    EncodeCoefficients BYVAL g_Comp(CompIndex(f)).ptrData, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  END IF
              NEXT h
            NEXT g
        NEXT f

        ' Advance .data pointers
        FOR f = FirstIndex TO SecondIndex
            FOR g = 0 TO g_Comp(CompIndex(f)).Vi - 1
              p(f, g) = p(f, g) + pLF(f)
            NEXT g
        NEXT f
      NEXT j

      ' Encode Bottommost MCU Scanline
      FOR i = 1 TO MCUx
        FOR f = FirstIndex TO SecondIndex
            FOR g = 1 TO g_Comp(CompIndex(f)).Vi
              FOR h = 1 TO g_Comp(CompIndex(f)).vHi
                  IF p(f, g - 1) > piaUBOUND(BYVAL g_Comp(CompIndex(f)).ptrData) OR (i = MCUx AND h > MCUr(f)) THEN
                    ' Pad with dummy block
                    Pad64(0) = Pred(f)
                    EncodeCoefficients Pad64(), 0, Pred(f), Td(f), Ta(f)
                  ELSE
                    EncodeCoefficients BYVAL g_Comp(CompIndex(f)).ptrData, p(f, g - 1), Pred(f), Td(f), Ta(f)
                  END IF
              NEXT h
            NEXT g
        NEXT f
      NEXT i

      WriteBitsEnd
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error21",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------


SUB InsertSequentialScans(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
  BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG)
  TRY
      ' Insert scan components CompIndex(FirstIndex) to CompIndex(SecondIndex) sequentially in compliance
      ' with JPEG rules. Components are interleaved whenever possible to emit as few scans as possible.
      LOCAL f    AS LONG    ' First Index
      LOCAL g    AS LONG    ' Second Index
      LOCAL nb  AS LONG    ' Number of 8X8 blocks in MCU
      LOCAL flag AS LONG    ' True when ready to insert scan(s)

      f = FirstIndex
      g = FirstIndex
      nb = 0
      flag = %FALSE

      DO WHILE f <= SecondIndex
        nb = nb + g_Comp(CompIndex(g)).vHi * g_Comp(CompIndex(g)).Vi
        g = g + 1

        IF nb > 10 THEN                '%MAXNB    = 10        ' Max 8X8 blocks in MCU  (10 for JPEG compliance)
            flag = %TRUE
            IF f <> g - 1 THEN g = g - 1
        ELSE
            IF (g - f) = 3 OR g > SecondIndex THEN flag = %TRUE
        END IF

        IF flag THEN
            IF f = g - 1 THEN
              InsertSOSNonInterleaved CompIndex(f), Td(f), Ta(f)
            ELSE
              InsertSOSInterleaved CompIndex(), Td(), Ta(), f, g - 1
            END IF

            nb = 0
            f = g
            flag = %FALSE
        END IF
      LOOP
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error22",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Writing file
'------------------------------------------------------------------------------------
FUNCTION OptimizeHuffmanTables(BYREF CompIndex() AS LONG, BYREF Td() AS LONG, BYREF Ta() AS LONG, _
  BYVAL FirstIndex AS LONG, BYVAL SecondIndex AS LONG) AS LONG
  TRY
      ' Optimize Huffman tables for the component indexes given.
      ' Returns an estimate of the number of bytes needed for entropy coded data.
      ' Estimate assumes a single scan, and entropy coded FF bytes are not followed by a zero stuff byte.
      LOCAL f        AS LONG  ' First Index
      LOCAL g        AS LONG  ' Second Index
      LOCAL i        AS LONG
      LOCAL j        AS LONG
      LOCAL k        AS LONG
      LOCAL CodeLen  AS LONG  ' Total bytes required for entropy coded data
      LOCAL nb        AS LONG  ' Number of 8X8 blocks in MCU


      DIM  Freq(256) AS LONG  ' Frequency count for optimizing Huffman tables
      LOCAL Freq2()  AS LONG  ' Copy of freq, used for calcultating entropy coded data size
      LOCAL IsInter() AS LONG  ' True if component i will be interleaved
      LOCAL TdUsed()  AS LONG  ' True if g_HuffDC(i) is used
      LOCAL TaUsed()  AS LONG  ' True if g_HuffAC(i) is used
      LOCAL flag      AS LONG  ' True when ready to include scan(s)

      REDIM IsInter(FirstIndex TO SecondIndex)
      REDIM TaUsed(3)
      REDIM TdUsed(3)

      ' Determine which components will be interleaved by InsertSequentialScans(), which tables are used
      f = FirstIndex
      g = FirstIndex
      nb = 0
      flag = %FALSE

      DO WHILE f <= SecondIndex
        nb = nb + g_Comp(CompIndex(g)).vHi * g_Comp(CompIndex(g)).Vi
        g = g + 1

        IF nb > 10 THEN        '%MAXNB          = 10      ' Max 8X8 blocks in MCU  (10 for JPEG compliance)
            flag = %TRUE
            IF f <> g - 1 THEN g = g - 1
        ELSE
            IF (g - f) = 3 OR g > SecondIndex THEN flag = %TRUE
        END IF

        IF flag THEN
            IF f = g - 1 THEN
              TdUsed(Td(f)) = %TRUE
              TaUsed(Ta(f)) = %TRUE
              IsInter(f) = %FALSE
            ELSE
              FOR i = f TO g - 1
                  TdUsed(Td(i)) = %TRUE
                  TaUsed(Ta(i)) = %TRUE
                  IsInter(i) = %TRUE
              NEXT i
            END IF

            nb = 0
            f = g
            flag = %FALSE
        END IF
      LOOP

      ' Optimize huffman tables for the scan sequence
      CodeLen = 0

      FOR i = 0 TO 3
        IF TdUsed(i) THEN
            FOR f = FirstIndex TO SecondIndex
              IF Td(f) = i THEN
                  IF IsInter(f) THEN
                    CALL CollectStatisticsDCInterleaved(BYVAL g_Comp(CompIndex(f)).ptrData, Freq(), _
                    g_Comp(CompIndex(f)).vHi, g_Comp(CompIndex(f)).Vi)
                  ELSE
                    CALL CollectStatisticsDCNonInterleaved(BYVAL g_Comp(CompIndex(f)).ptrData, Freq())
                  END IF
              END IF
            NEXT f

            ' Optimize and create this DC table
            'Freq2() = Freq()
            REDIM Freq2(UBOUND(Freq))

            FOR k = 0 TO UBOUND(Freq)
              Freq2(k) = Freq(k)
            NEXT k

            CALL BuildHuffman(g_HuffDC(i), Freq())
            CALL ExpandHuffman(g_HuffDC(i), IIF(g_PP = 12, 15, 11))
            CodeLen = CodeLen + CodeLength(g_HuffDC(i), Freq2())
        END IF

        IF TaUsed(i) THEN
            FOR f = FirstIndex TO SecondIndex
              IF Td(f) = i THEN CALL CollectStatisticsAC(BYVAL g_Comp(CompIndex(f)).ptrData, Freq())
            NEXT f

            ' Optimize and create this AC table
            'Freq2() = Freq()
            REDIM Freq2(UBOUND(Freq))

            FOR k = 0 TO UBOUND(Freq)
              Freq2(k) = Freq(k)
            NEXT k

            CALL BuildHuffman(g_HuffAC(i), Freq())
            CALL ExpandHuffman(g_HuffAC(i), 255)
            CodeLen = CodeLen + CodeLength(g_HuffAC(i), Freq2())
        END IF
      NEXT i

      FUNCTION = CodeLen
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error23",,ERROR$(ERR)
  END TRY
END FUNCTION

'------------------------------------------------------------------------------------

FUNCTION PBJPEG_SaveFile(FileName AS STRING) AS LONG
  TRY
      DIM CompIndex() AS LONG    ' Indexes of Components to be included
      DIM Td()        AS LONG    ' DC Huffman Table Selectors
      DIM Ta()        AS LONG    ' AC Huffman Table Selectors
      DIM FileNum    AS INTEGER
      DIM i          AS LONG

      IF LEN(FileName) = 0 THEN
        FUNCTION = 1          ' FileName not given
      ELSE
        IF (LEN(DIR$(FileName, %NORMAL OR %READONLY OR %HIDDEN OR %SYSTEM OR %ARCHIVE)) > 0) THEN
            FUNCTION = 2      ' File already exists
        ELSE
            REDIM CompIndex(g_Nf - 1)
            REDIM Td(g_Nf - 1)
            REDIM Ta(g_Nf - 1)

            FOR i = 0 TO g_Nf - 1
              CompIndex(i) = i
              Td(i) = IIF(i = 0, 0, 1)
              Ta(i) = IIF(i = 0, 0, 1)
            NEXT i

            i = OptimizeHuffmanTables(CompIndex(), Td(), Ta(), 0, g_Nf - 1)

            ' Estimate maximum possible file size needed
            REDIM g_m_Data(i * 2 + 1000 + LEN(g_m_Comment))
            g_m_Ptr = 0

            InsertMarker %SOI                                  ' SOI - Start of Image
            InsertJFIF                                        ' JFIF

            IF LEN(g_m_Comment) > 0 THEN InsertCOM g_m_Comment    ' COM - Comment
            InsertCOM "PowerBasic JPEG Encoder by Philipp E. Weidmann" & $CRLF _
            & "" & $CRLF _
            & "Ported & adapted from cJPEGI by Ron van Tilbug and cJPEG by John Korejwa"

            InsertDQT g_m_Ptr, 0                                ' DQT - Define Quantization Tables
            IF g_Nf > 1 THEN InsertDQT g_m_Ptr, 1

            InsertSOF %SOF0                                    ' SOF - Start of Frame

            InsertDHT g_m_Ptr, 0, %FALSE                        ' DHT - Define Huffman Tables
            InsertDHT g_m_Ptr, 0, %TRUE

            IF g_Nf > 1 THEN
              InsertDHT g_m_Ptr, 1, %FALSE
              InsertDHT g_m_Ptr, 1, %TRUE
            END IF

            InsertSequentialScans CompIndex(), Td(), Ta(), 0, g_Nf - 1  ' SOS - Scan Data
            InsertMarker %EOI                                        ' EOI - End of Image

            ' Size the final byte array and write to file
            REDIM PRESERVE g_m_Data(g_m_Ptr - 1)
            FileNum = FREEFILE

            OPEN FileName FOR BINARY ACCESS WRITE AS FileNum
            PUT #FileNum, , g_m_Data()
            CLOSE FileNum

            ERASE g_m_Data
        END IF
      END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error24",,ERROR$(ERR)
  END TRY
END FUNCTION


SUB PBJPEG_Reset()
  'free up memory
  TRY
      REDIM g_QLumin(1)
      REDIM g_QChrom(1)
      REDIM g_TwoP(1)
      REDIM g_TwoR(1)
      REDIM g_TwoB(1)
      REDIM g_EIDCTScale(1)
      REDIM g_DIDCTScale(1)
      REDIM g_ZigZag(1)
      g_PP      = 0    ' Sample Precision [8, 12]
      g_YY      = 0    ' Number of lines            [Image Height] after clipping
      g_XX      = 0    ' Number of samples per line  [Image Width]  after clipping
      g_Nf      = 0    ' Number of components in Frame

      g_HMax    = 0    ' Maximum horizontal sampling frequency
      g_VMax    = 0    ' Maximum vertical sampling frequency

      REDIM g_m_Data(1)
      g_m_Chr    = 0    ' Current Character in g_m_Data
      g_m_Ptr    = 0    ' Byte index in g_m_Data
      g_m_Bit    = 0    ' Bit index in g_m_Chr

      REDIM g_m_Block(1)

      REDIM g_QTable(1)
      REDIM g_HuffDC(1)
      REDIM g_HuffAC(1)
      REDIM g_Comp(1)

      REDIM g_CompData0(1)
      REDIM g_CompData1(1)
      REDIM g_CompData2(1)

      g_m_Quality  = 0
      g_m_Comment  = ""

      '      Dim g_QLumin(63)      As Global Byte
      '      Dim g_QChrom(63)      As Global Byte
      '      Dim g_TwoP(31)        As Global Long
      '      Dim g_TwoR(31)        As Global Long
      '      Dim g_TwoB(255)        As Global Byte
      '      Dim g_EIDCTScale(7, 7) As Global Long
      '      Dim g_DIDCTScale(7, 7) As Global Long
      '      Dim g_ZigZag(7, 7)    As Global Long
      '      Dim g_m_Block(7, 7)    As Global Long
      '      Dim g_QTable(3)        As Global QUANTIZATIONTABLE
      '      Dim g_HuffDC(3)        As Global HUFFMANTABLE
      '      Dim g_HuffAC(3)        As Global HUFFMANTABLE
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error25a",,ERROR$(ERR)
  END TRY
END SUB


'------------------------------------------------------------------------------------
' Initialization
'------------------------------------------------------------------------------------
FUNCTION PBJPEG_Initialize() AS LONG
  TRY
      LOCAL i  AS LONG
      LOCAL j  AS LONG
      LOCAL dx AS LONG
      LOCAL zz AS LONG

      ' Dimension global arrays
      DIM g_QLumin(63)      AS GLOBAL BYTE
      DIM g_QChrom(63)      AS GLOBAL BYTE
      DIM g_TwoP(31)        AS GLOBAL LONG
      DIM g_TwoR(31)        AS GLOBAL LONG
      DIM g_TwoB(255)        AS GLOBAL BYTE
      DIM g_EIDCTScale(7, 7) AS GLOBAL LONG
      DIM g_DIDCTScale(7, 7) AS GLOBAL LONG
      DIM g_ZigZag(7, 7)    AS GLOBAL LONG
      DIM g_m_Block(7, 7)    AS GLOBAL LONG
      DIM g_QTable(3)        AS GLOBAL QUANTIZATIONTABLE
      DIM g_HuffDC(3)        AS GLOBAL HUFFMANTABLE
      DIM g_HuffAC(3)        AS GLOBAL HUFFMANTABLE

      i = 0            ' Initialize the g_ZigZag() array, which maps out the
      j = 0            ' zig-zag sequence of quantized DCT coefficients
      dx = 1          ' in approximately low to high spatial frequencies

      FOR zz = 0 TO 63
        g_ZigZag(i, j) = zz
        i = i + dx
        j = j - dx

        IF i > 7 THEN          '  0  1  5  6  14  15  27  28
            i = 7              '  2  4  7  13  16  26  29  42
            j = j + 2          '  3  8  12  17  25  30  41  43
            dx = -1            '  9  11  18  24  31  40  44  53
        ELSEIF j > 7 THEN      ' 10  19  23  32  39  45  52  54
            j = 7              ' 20  22  33  38  46  51  55  60
            i = i + 2          ' 21  34  37  47  50  56  59  61
            dx = 1            ' 35  36  48  49  57  58  62  63
        ELSEIF i < 0 THEN
            i = 0              ' Check (j>7) first
            dx = 1
        ELSEIF j < 0 THEN
            j = 0
            dx = -1
        END IF
      NEXT zz

      ' Luminance Quantization table for Quality = 50
      g_QLumin(0) = 16:  g_QLumin(1) = 11:  g_QLumin(2) = 12:  g_QLumin(3) = 14
      g_QLumin(4) = 12:  g_QLumin(5) = 10:  g_QLumin(6) = 16:  g_QLumin(7) = 14
      g_QLumin(8) = 13:  g_QLumin(9) = 14:  g_QLumin(10) = 18:  g_QLumin(11) = 17
      g_QLumin(12) = 16:  g_QLumin(13) = 19:  g_QLumin(14) = 24:  g_QLumin(15) = 40
      g_QLumin(16) = 26:  g_QLumin(17) = 24:  g_QLumin(18) = 22:  g_QLumin(19) = 22
      g_QLumin(20) = 24:  g_QLumin(21) = 49:  g_QLumin(22) = 35:  g_QLumin(23) = 37
      g_QLumin(24) = 29:  g_QLumin(25) = 40:  g_QLumin(26) = 58:  g_QLumin(27) = 51
      g_QLumin(28) = 61:  g_QLumin(29) = 60:  g_QLumin(30) = 57:  g_QLumin(31) = 51
      g_QLumin(32) = 56:  g_QLumin(33) = 55:  g_QLumin(34) = 64:  g_QLumin(35) = 72
      g_QLumin(36) = 92:  g_QLumin(37) = 78:  g_QLumin(38) = 64:  g_QLumin(39) = 68
      g_QLumin(40) = 87:  g_QLumin(41) = 69:  g_QLumin(42) = 55:  g_QLumin(43) = 56
      g_QLumin(44) = 80:  g_QLumin(45) = 109: g_QLumin(46) = 81:  g_QLumin(47) = 87
      g_QLumin(48) = 95:  g_QLumin(49) = 98:  g_QLumin(50) = 103: g_QLumin(51) = 104
      g_QLumin(52) = 103: g_QLumin(53) = 62:  g_QLumin(54) = 77:  g_QLumin(55) = 113
      g_QLumin(56) = 121: g_QLumin(57) = 112: g_QLumin(58) = 100: g_QLumin(59) = 120
      g_QLumin(60) = 92:  g_QLumin(61) = 101: g_QLumin(62) = 103: g_QLumin(63) = 99

      ' Chrominance Quantization table for Quality = 50
      g_QChrom(0) = 17:  g_QChrom(1) = 18:  g_QChrom(2) = 18:  g_QChrom(3) = 24
      g_QChrom(4) = 21:  g_QChrom(5) = 24:  g_QChrom(6) = 47:  g_QChrom(7) = 26
      g_QChrom(8) = 26:  g_QChrom(9) = 47:  g_QChrom(10) = 99:  g_QChrom(11) = 66
      g_QChrom(12) = 56:  g_QChrom(13) = 66:  g_QChrom(14) = 99:  g_QChrom(15) = 99
      g_QChrom(16) = 99:  g_QChrom(17) = 99:  g_QChrom(18) = 99:  g_QChrom(19) = 99
      g_QChrom(20) = 99:  g_QChrom(21) = 99:  g_QChrom(22) = 99:  g_QChrom(23) = 99
      g_QChrom(24) = 99:  g_QChrom(25) = 99:  g_QChrom(26) = 99:  g_QChrom(27) = 99
      g_QChrom(28) = 99:  g_QChrom(29) = 99:  g_QChrom(30) = 99:  g_QChrom(31) = 99
      g_QChrom(32) = 99:  g_QChrom(33) = 99:  g_QChrom(34) = 99:  g_QChrom(35) = 99
      g_QChrom(36) = 99:  g_QChrom(37) = 99:  g_QChrom(38) = 99:  g_QChrom(39) = 99
      g_QChrom(40) = 99:  g_QChrom(41) = 99:  g_QChrom(42) = 99:  g_QChrom(43) = 99
      g_QChrom(44) = 99:  g_QChrom(45) = 99:  g_QChrom(46) = 99:  g_QChrom(47) = 99
      g_QChrom(48) = 99:  g_QChrom(49) = 99:  g_QChrom(50) = 99:  g_QChrom(51) = 99
      g_QChrom(52) = 99:  g_QChrom(53) = 99:  g_QChrom(54) = 99:  g_QChrom(55) = 99
      g_QChrom(56) = 99:  g_QChrom(57) = 99:  g_QChrom(58) = 99:  g_QChrom(59) = 99
      g_QChrom(60) = 99:  g_QChrom(61) = 99:  g_QChrom(62) = 99:  g_QChrom(63) = 99

      ' Encoding Quantization Scalars  (values are 0.. 0.5 scaled by 131072)
      ' Call GenEIDCT
      g_EIDCTScale(0, 0) = 16384: g_EIDCTScale(0, 1) = 16384: g_EIDCTScale(0, 2) = 25080: g_EIDCTScale(0, 3) = 23170
      g_EIDCTScale(0, 4) = 32768: g_EIDCTScale(0, 5) = 23170: g_EIDCTScale(0, 6) = 21407: g_EIDCTScale(0, 7) = 32768
      g_EIDCTScale(1, 0) = 16384: g_EIDCTScale(1, 1) = 16384: g_EIDCTScale(1, 2) = 25080: g_EIDCTScale(1, 3) = 23170
      g_EIDCTScale(1, 4) = 32768: g_EIDCTScale(1, 5) = 23170: g_EIDCTScale(1, 6) = 21407: g_EIDCTScale(1, 7) = 32768
      g_EIDCTScale(2, 0) = 25080: g_EIDCTScale(2, 1) = 25080: g_EIDCTScale(2, 2) = 38390: g_EIDCTScale(2, 3) = 35468
      g_EIDCTScale(2, 4) = 50159: g_EIDCTScale(2, 5) = 35468: g_EIDCTScale(2, 6) = 32768: g_EIDCTScale(2, 7) = 50159
      g_EIDCTScale(3, 0) = 23170: g_EIDCTScale(3, 1) = 23170: g_EIDCTScale(3, 2) = 35468: g_EIDCTScale(3, 3) = 32768
      g_EIDCTScale(3, 4) = 46341: g_EIDCTScale(3, 5) = 32768: g_EIDCTScale(3, 6) = 30274: g_EIDCTScale(3, 7) = 46341
      g_EIDCTScale(4, 0) = 32768: g_EIDCTScale(4, 1) = 32768: g_EIDCTScale(4, 2) = 50159: g_EIDCTScale(4, 3) = 46341
      g_EIDCTScale(4, 4) = 65536: g_EIDCTScale(4, 5) = 46341: g_EIDCTScale(4, 6) = 42813: g_EIDCTScale(4, 7) = 65536
      g_EIDCTScale(5, 0) = 23170: g_EIDCTScale(5, 1) = 23170: g_EIDCTScale(5, 2) = 35468: g_EIDCTScale(5, 3) = 32768
      g_EIDCTScale(5, 4) = 46341: g_EIDCTScale(5, 5) = 32768: g_EIDCTScale(5, 6) = 30274: g_EIDCTScale(5, 7) = 46341
      g_EIDCTScale(6, 0) = 21407: g_EIDCTScale(6, 1) = 21407: g_EIDCTScale(6, 2) = 32768: g_EIDCTScale(6, 3) = 30274
      g_EIDCTScale(6, 4) = 42813: g_EIDCTScale(6, 5) = 30274: g_EIDCTScale(6, 6) = 27969: g_EIDCTScale(6, 7) = 42813
      g_EIDCTScale(7, 0) = 32768: g_EIDCTScale(7, 1) = 32768: g_EIDCTScale(7, 2) = 50159: g_EIDCTScale(7, 3) = 46341
      g_EIDCTScale(7, 4) = 65536: g_EIDCTScale(7, 5) = 46341: g_EIDCTScale(7, 6) = 42813: g_EIDCTScale(7, 7) = 65536

      ' Powers of 2
      CALL Init_BitMasks

      ' The number of bits in value
      g_TwoB(0) = 0: g_TwoB(1) = 1: g_TwoB(2) = 2: g_TwoB(3) = 2

      FOR i = 4 TO 7
        g_TwoB(i) = 3
      NEXT i
      FOR i = 8 TO 15
        g_TwoB(i) = 4
      NEXT i
      FOR i = 16 TO 31
        g_TwoB(i) = 5
      NEXT i
      FOR i = 32 TO 63
        g_TwoB(i) = 6
      NEXT i
      FOR i = 64 TO 127
        g_TwoB(i) = 7
      NEXT i
      FOR i = 128 TO 255
        g_TwoB(i) = 8
      NEXT i

      PBJPEG_SetSamplingFrequencies 2, 2, 1, 1, 1, 1
      PBJPEG_SetQuality 75
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error25",,ERROR$(ERR)
  END TRY
END FUNCTION

'------------------------------------------------------------------------------------

SUB Init_BitMasks()
  TRY
      LOCAL i      AS LONG
      DIM  sb(31) AS LONG

      '    Requires PB 8
      '    ARRAY ASSIGN sb() = &H1&, &H2&, &H4&, &H8&, _
      '                        &H10&, &H20&, &H40&, &H80&, _
      '                        &H100&, &H200&, &H400&, &H800&, _
      '                        &H1000&, &H2000&, &H4000, &H8000&, _
      '                        &H10000, &H20000, &H40000, &H80000, _
      '                        &H100000, &H200000, &H400000, &H800000, _
      '                        &H1000000, &H2000000, &H4000000, &H8000000, _
      '                        &H10000000, &H20000000, &H40000000, &H80000000
      sb(0)  = &H1&
      sb(1)  = &H2&
      sb(2)  = &H4&
      sb(3)  = &H8&
      sb(4)  = &H10&
      sb(5)  = &H20&
      sb(6)  = &H40&
      sb(7)  = &H80&
      sb(8)  = &H100&
      sb(9)  = &H200&
      sb(10)  = &H400&
      sb(11)  = &H800&
      sb(12)  = &H1000&
      sb(13)  = &H2000&
      sb(14)  = &H4000
      sb(15)  = &H8000&
      sb(16)  = &H10000
      sb(17)  = &H20000
      sb(18)  = &H40000
      sb(19)  = &H80000
      sb(20)  = &H100000
      sb(21)  = &H200000
      sb(22)  = &H400000
      sb(23)  = &H800000
      sb(24)  = &H1000000
      sb(25)  = &H2000000
      sb(26)  = &H4000000
      sb(27)  = &H8000000
      sb(28)  = &H10000000
      sb(29)  = &H20000000
      sb(30)  = &H40000000
      sb(31)  = &H80000000

      FOR i = 0 TO 30
        g_TwoP(i) = sb(i)
        g_TwoR(i + 1) = sb(i)
      NEXT i

      g_TwoP(31) = sb(31)
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error27",,ERROR$(ERR)
  END TRY
END SUB

'------------------------------------------------------------------------------------

' Uncomment this to generate the table above
SUB GenEIDCT()
  '    DIM  t(7) AS DOUBLE
  '    LOCAL i    AS LONG, j AS LONG
  '
  '    %PI = 3.14159265358979
  '
  '    t(0) = SIN(%PI / 4) / 2
  '    t(1) = 1 / SQR(8)
  '    t(2) = 1 / (2 * SIN(3 * %PI / 8))
  '    t(3) = 0.5
  '    t(4) = SIN(%PI / 4)
  '    t(5) = 0.5
  '    t(6) = SIN(3 * %PI / 8) / 2
  '    t(7) = 1 / SQR(2)
  '
  '    OPEN "EIDCT.txt" FOR OUTPUT AS #1
  '
  '    FOR i = 0 TO 7
  '        FOR j = 0 TO 7
  '            g_EIDCTScale(i, j) = 131072 * t(i) * t(j)
  '            PRINT #1, "g_EIDCTScale(" & i & "," & j & ")=" & g_EIDCTScale(i, j)
  '        NEXT j
  '    NEXT i
  '
  '    CLOSE #1
END SUB

'------------------------------------------------------------------------------------

' The routine below is functionally equivalent to that below but entirely in integers only
' nBits  = Int((Log(Abs(v)) * 1.442695040889))+1  '1/log(2) v<>0, 0 if v=0
FUNCTION nBits(BYVAL v AS LONG) AS LONG    ' Effectively binary search for Log2 -32768<v<32767
  TRY

      IF v < 0 THEN v = -v

      IF v < %TWOE08 THEN                          ' Use linear table lookup for values <256
        nBits = g_TwoB(v)
      ELSE                                        ' Use binary search
        '            IF v < %TWOE08 THEN                ' 0..255 bits 7..0
        '                IF v < %TWOE04 THEN            ' 0.. 15 bits 3..0
        '                    IF v < %TWOE02 THEN        ' 0..  1 bits 1..0
        '                        IF v < %TWOE01 THEN
        '                            nBits = 1
        '                        ELSE
        '                            nBits = 2
        '                        END IF
        '                    ELSEIF v < %TWOE03 THEN    ' 4..  7 bits 3..2
        '                        nBits = 3
        '                    ELSE
        '                        nBits = 4
        '                    END IF
        '                ELSE                            ' 16..255  bits 7..4
        '
        '                IF v < %TWOE06 THEN            ' 16.. 63  bits 5..4
        '                    IF v < %TWOE05 THEN
        '                        nBits = 5
        '                    ELSE
        '                        nBits = 6
        '                    END IF
        '                ELSEIF v < %TWOE07 THEN        ' 64..127 bits 7..6
        '                    nBits = 7
        '                ELSE
        '                    nBits = 8
        '                END IF
        '            END IF
        '        ELSE                                    ' 256..32768 bits 15..8
        IF v < %TWOE12 THEN                      ' Bits 11..8
            IF v < %TWOE10 THEN                  ' Bits  9..8
              IF v < %TWOE09 THEN
                  nBits = 9
              ELSE
                  nBits = 10
              END IF
            ELSEIF v < %TWOE11 THEN
              nBits = 11
            ELSE                                ' Bits 11..10
              nBits = 12
            END IF
        ELSE                                    ' Bits 15..12
            IF v < %TWOE14 THEN                  ' Bits 13..12
              IF v < %TWOE13 THEN
                  nBits = 13
              ELSE
                  nBits = 14
              END IF
            ELSEIF v < %TWOE15 THEN              ' Bits 15..14
              nBits = 15
            ELSE
              nBits = 16
            END IF
        END IF
      END IF
      '    END IF
  CATCH
      IF %NOMSG = 0 THEN MSGBOX "JPG Error28",,ERROR$(ERR)
  END TRY
END FUNCTION
